@@ -90,7 +90,7 @@ END
 `
 
 case "$cc" in
-'')    for i in `ls -r /opt/sol*studio*/bin/cc` /opt/SUNWspro/bin/cc
+'')    for i in `ls -r /opt/*studio*/bin/cc` /opt/SUNWspro/bin/cc
        do
 	       if test -f "$i"; then
 		       cc=$i
